Attribute VB_Name = "ModuleO"
'Dex3D
'A Visual Basic 3D Engine
'
'Copyright (C) 1999 by Jerry J. Chen
'
'This program is free software; you can redistribute it and/or
'modify it under the terms of the GNU General Public License
'as published by the Free Software Foundation; either version 2
'of the License, or (at your option) any later version.
'
'This program is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with this program; if not, write to the Free Software
'Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
'
'Jerry J. Chen
'onlyuser@hotmail.com

Type BITMAP
    bmType As Long 'bitmap type
    bmWidth As Long 'width of the bitmap in pixels
    bmHeight As Long 'height of the bitmap in raster lines
    bmWidthBytes As Long 'number of bytes in each raster line
    bmPlanes As Integer 'number of color planes in the bitmap
    bmBitsPixel As Integer 'number of adjacent color bits on each plane needed to define a pixel
    bmBits As Long 'pointer to the location of the bit values for the bitmap
End Type

Type SAFEARRAYBOUND
    cElements As Long 'lower bound of the dimension
    lLbound As Long 'number of elements in the dimension
End Type

Type SAFEARRAY
    cDims As Integer 'count of dimensions in this array
    fFeatures As Integer 'flags used by the SafeArray
    cbElements As Long 'size of an element of the array
    cLocks As Long 'number of times the array has been locked without corresponding unlock
    pvData As Long 'pointer to the data
    Bounds(0 To 1) As SAFEARRAYBOUND 'one bound for each dimension
End Type

Type TPalette
    Index As Byte
    Color As Long
End Type

Public Declare Function GetObject Lib "gdi32" Alias "GetObjectA" (ByVal hObject As Long, ByVal nCount As Long, lpObject As Any) As Long
Public Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (Destination As Any, Source As Any, ByVal Length As Long)
Public Declare Function VarPtrArray Lib "msvbvm50.dll" Alias "VarPtr" (Ptr() As Any) As Long
Public Declare Function GetPixel Lib "gdi32" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long) As Long

Public BMP As BITMAP
Public SA As SAFEARRAY
Public IndexByColor(6, 6, 6) As Byte
Public ColorByIndex(255) As ColorRGB
Public Pixel() As Byte
Function BuildColorTable(Canvas As PictureBox, ColorCount As Byte, Owner As Form)

    Dim A As Integer
    Dim B As Integer
    Dim C As Integer
    Dim D As Integer
    Dim VPalette() As TPalette
    Dim FoundMatch As Boolean
    Dim BufferA As ColorRGB
    Dim CurrentOffset As Single
    Dim BestOffset As Single
    
    ReDim VPalette(0)
    
    For A = 0 To 255
        Owner.Caption = "Populating palette: " & Int(A / 255 * 100) & "%"
        DoEvents
        Pixel(0, BMP.bmHeight - 1) = A
        Canvas.Cls
        For B = LBound(VPalette) To UBound(VPalette) - 1
            If VPalette(B).Index = Pixel(0, BMP.bmHeight - 1) Then
                FoundMatch = True
                Exit For
            End If
        Next B
        If FoundMatch = False Then
            VPalette(UBound(VPalette)).Index = A
            VPalette(UBound(VPalette)).Color = GetPixel(Canvas.hdc, 0, 0)
            ReDim Preserve VPalette(UBound(VPalette) + 1)
        Else
            FoundMatch = False
        End If
    Next A
    Pixel(0, BMP.bmHeight - 1) = 0
    
    For A = 0 To 6
        For B = 0 To 6
            For C = 0 To 6
                Owner.Caption = _
                    "Building color table: " & _
                    Int( _
                        ( _
                            A / 7 ^ 1 + _
                            B / 7 ^ 2 + _
                            C / 7 ^ 3 _
                        ) * 100 _
                    ) & "%"
                DoEvents
                BestOffset = 0
                For D = LBound(VPalette) To LBound(VPalette) + ColorCount - 1
                    BufferA = ColorLongToRGB(VPalette(D).Color)
                    CurrentOffset = _
                        VectorDistance( _
                            VectorInput( _
                                Val(BufferA.R), _
                                Val(BufferA.G), _
                                Val(BufferA.B) _
                            ), _
                            VectorInput( _
                                Val(A) * 255 / 6, _
                                Val(B) * 255 / 6, _
                                Val(C) * 255 / 6 _
                            ) _
                        )
                    If BestOffset = 0 Or CurrentOffset < BestOffset Then
                        BestOffset = CurrentOffset
                        IndexByColor(A, B, C) = VPalette(D).Index
                        ColorByIndex(VPalette(D).Index) = BufferA
                    End If
                Next D
            Next C
        Next B
    Next A
    
    Owner.Caption = "Done!"
    
End Function
Function CanvasBox(X1 As Integer, Y1 As Integer, X2 As Integer, Y2 As Integer, Color As Byte)

    Dim Y As Integer
    Dim IncY As Integer
    
    If Y1 < Y2 Then
        IncY = 1
    Else
        IncY = -1
    End If
    
    For Y = Y1 To Y2 Step IncY
        Call CanvasScanLine(Y, X1, X2, Color)
    Next Y
    
End Function
Function CanvasFill(Color As Byte)

    Call _
        CanvasBox( _
            0, 0, _
            BMP.bmWidth - 1, BMP.bmHeight - 1, _
            Color _
        )
        
End Function
Function CanvasLine(X1 As Integer, Y1 As Integer, X2 As Integer, Y2 As Integer, Color As Byte)

    Dim DeltaX As Integer
    Dim DeltaY As Integer
    Dim IncX As Integer
    Dim IncY As Integer
    Dim X As Integer
    Dim Y As Integer
    Dim DetX As Integer
    Dim DetY As Integer
    
    DeltaX = X2 - X1
    DeltaY = Y2 - Y1
    
    If X2 > X1 Then
        IncX = 1
    ElseIf X2 < X1 Then
        IncX = -1
    Else
        IncX = 0
    End If
    
    If Y2 > Y1 Then
        IncY = 1
    ElseIf Y2 < Y1 Then
        IncY = -1
    Else
        IncY = 0
    End If
    
    X = X1
    Y = Y1
    
    Do
    
        Pixel(X, Y) = Color
        
        DetX = DeltaY * (X - X1)
        DetY = DeltaX * (Y - Y1)
        
        If DetX <> DetY Then
            If IncX = IncY Then
                If DetX > DetY Then
                    Y = Y + IncY
                ElseIf DetX < DetY Then
                    X = X + IncX
                End If
            Else
                If DetX < DetY Then
                    Y = Y + IncY
                ElseIf DetX > DetY Then
                    X = X + IncX
                End If
            End If
        Else
            X = X + IncX
            Y = Y + IncY
        End If
        
    Loop Until X = X2 And Y = Y2
    
End Function
Function CanvasPixel(X As Integer, Y As Integer, Color As Byte)

    Pixel(X, Y) = Color
    
End Function
Function CanvasScanLine(Y As Integer, X1 As Integer, X2 As Integer, Color As Byte)

    Dim X As Integer
    Dim IncX As Integer
    
    If X1 < X2 Then
        IncX = 1
    Else
        IncX = -1
    End If
    
    For X = X1 To X2 Step IncX
        Pixel(X, Y) = Color
    Next X
    
End Function
Function CanvasTriangle(Canvas As PictureBox, A As POINTAPI, B As POINTAPI, C As POINTAPI, Color As Byte)

    Dim MaxY As Integer
    Dim MinY As Integer
    Dim MidY As Integer
    Dim X1 As Integer
    Dim X2 As Integer
    Dim FootA As POINTAPI
    Dim FootB As POINTAPI
    Dim PeakA As POINTAPI
    Dim PeakB As POINTAPI
    Dim Y As Integer
    Dim IncY As Integer
    
    MaxY = A.Y
    If B.Y > MaxY Then MaxY = B.Y
    If C.Y > MaxY Then MaxY = C.Y
    
    MinY = A.Y
    If B.Y < MinY Then MinY = B.Y
    If C.Y < MinY Then MinY = C.Y
    
    If A.Y <> MaxY And A.Y <> MinY Then MidY = A.Y
    If B.Y <> MaxY And B.Y <> MinY Then MidY = B.Y
    If C.Y <> MaxY And C.Y <> MinY Then MidY = C.Y
    
    If MaxY = MinY Then 'horizontal line
        X1 = A.X
        If B.X > X1 Then X1 = B.X
        If C.X > X1 Then X1 = C.X
        X2 = A.X
        If B.X < X2 Then X2 = B.X
        If C.X < X2 Then X2 = C.X
        Call CanvasScanLine(MaxY, X1, X2, Color)
    Else 'triangle
        If MidY = 0 Then 'horizontal base (one sub-triangle)
            If A.Y = B.Y Then
                FootA = A
                FootB = B
                PeakA = C
            ElseIf B.Y = C.Y Then
                FootA = B
                FootB = C
                PeakA = A
            Else
                FootA = C
                FootB = A
                PeakA = B
            End If
            PeakB = PeakA
        Else 'tilted base (two sub-triangles)
            If A.Y = MidY Then
                FootA = A
                PeakA = B
                PeakB = C
            ElseIf B.Y = MidY Then
                FootA = B
                PeakA = C
                PeakB = A
            Else
                FootA = C
                PeakA = A
                PeakB = B
            End If
            FootB.X = PeakB.X + (FootA.Y - PeakB.Y) / (PeakA.Y - PeakB.Y) * (PeakA.X - PeakB.X)
            FootB.Y = FootA.Y
            If PeakB.Y < FootA.Y Then
                IncY = 1
            Else
                IncY = -1
            End If
            For Y = PeakB.Y To FootA.Y - IncY Step IncY
                X1 = PeakB.X + (FootA.X - PeakB.X) / (FootA.Y - PeakB.Y) * (Y - PeakB.Y)
                X2 = PeakB.X + (FootB.X - PeakB.X) / (FootB.Y - PeakB.Y) * (Y - PeakB.Y)
                Call CanvasScanLine(Y, X1, X2, Color)
            Next Y
        End If
        If PeakA.Y < FootA.Y Then
            IncY = 1
        Else
            IncY = -1
        End If
        For Y = PeakA.Y To FootA.Y Step IncY
            X1 = PeakA.X + (FootA.X - PeakA.X) / (FootA.Y - PeakA.Y) * (Y - PeakA.Y)
            X2 = PeakA.X + (FootB.X - PeakA.X) / (FootB.Y - PeakA.Y) * (Y - PeakA.Y)
            Call CanvasScanLine(Y, X1, X2, Color)
        Next Y
    End If
    
End Function
Function GetColorIndex(Color As Long) As Byte

    Dim BufferA As ColorRGB
    
    BufferA = ColorScale(ColorLongToRGB(Color), 6 / 255)
    GetColorIndex = _
        IndexByColor( _
            BufferA.R, _
            BufferA.G, _
            BufferA.B _
        )
        
End Function
Function KillSafeArray()

    Call CopyMemory(ByVal VarPtrArray(Pixel), CLng(0), 4)
    
End Function
Function InitializeSafeArray(Canvas As PictureBox, Filename As String)

    If Filename <> "" Then Canvas.Picture = LoadPicture(Filename)
    Call GetObject(Canvas.Picture, Len(BMP), BMP)
    If BMP.bmPlanes = 1 And BMP.bmBitsPixel = 8 Then
        SA.cbElements = 1
        SA.cDims = 2
        SA.Bounds(0).lLbound = 0
        SA.Bounds(0).cElements = BMP.bmHeight
        SA.Bounds(1).lLbound = 0
        SA.Bounds(1).cElements = BMP.bmWidthBytes
        SA.pvData = BMP.bmBits
        Call CopyMemory(ByVal VarPtrArray(Pixel), VarPtr(SA), 4)
    Else
        Call MsgBox("256-color bitmaps only!", 16, "Error")
    End If
    
End Function
Sub CanvasTest()

    Dim Y As Integer
    
    For Y = 0 To BMP.bmHeight - 1
        Call CanvasScanLine(Y, 0, BMP.bmWidth - 1, Y / (BMP.bmHeight - 1) * 255)
    Next Y
    
End Sub
